/* jslint node:true */
/* global describe:true */
/* global it:true */

'use strict';

const expect = require('expect.js'),
    manifestFormat = require('../manifestformat.js'),
    _ = require('underscore');

describe('parseManifest', function () {
    it('errors for empty string', function () {
        expect(manifestFormat.parseString('').error).to.be.an(Error);
    });

    it('errors for invalid json', function () {
        expect(manifestFormat.parseString('garbage').error).to.be.an(Error);
    });

    const manifest = {
        id: 'io.cloudron.test',
        title: 'Bar App',
        description: 'Long long ago, there was foo',
        tagline: 'Not your usual Foo App',
        author: 'Herbert Burgermeister',
        manifestVersion: 1,
        version: '0.1.2',
        dockerImage: 'girish/foo:0.2',
        httpPort: 23,
        website: 'https://example.com',
        contactEmail: 'support@example.com'
    };

    manifestFormat.SCHEMA_V1.required.forEach(function (key) {
        const manifestCopy = _.extend({ }, manifest);
        delete manifestCopy[key];
        it('errors for missing ' + key, function () {
            expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
        });
    });

    new Array(null, [ 23 ], [ 'mysql', 34 ], [ null, 'mysql' ]).forEach(function (invalidAddon, idx) {
        it('fails for invalid addon testcase ' + idx, function () {
            const manifestCopy = _.extend({ }, manifest);
            manifestCopy.addons = invalidAddon;
            expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
        });
    });

    it('fails for bad version', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.version = '0.2';
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('fails for bad id', function () {
        const manifestCopy = _.extend({ }, manifest);
        ['simply', '12de', 'in..x', 'x.com', 'no.hy-phen' ].forEach(function(badId) {
            manifestCopy.id = badId;
            expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
        });
    });

    it('fails for bad minBoxVersion', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.minBoxVersion = '0.2';
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('fails for bad maxBoxVersion', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.maxBoxVersion = '0.2';
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('fails for bad targetBoxVersion', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.targetBoxVersion = '0.2';
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('fails for bad manifestVersion', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.manifestVersion = 3;
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('fails for bad tcpPorts', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.tcpPorts = 45;

        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('fails for bad udpPorts', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.udpPorts = 45;

        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('fails for bad capabilities', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.capabilities = [ 'test' ];

        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('fails for bad changelog', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.changelog = 34;

        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);

        manifestCopy.changelog = [ ];
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);

        manifestCopy.changelog = [ null ];
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);

        manifestCopy.changelog = [ 56 ];
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('fails for bad tcpPorts', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.tcpPorts = 45;
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);

        manifestCopy.tcpPorts = { 'env$': { } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad env

        manifestCopy.tcpPorts = { 'CLOUDRON_OOPS': { } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad env

        manifestCopy.tcpPorts = { 'env': { } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // missing description

        manifestCopy.tcpPorts = { 'env': { description: 34 } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad description

        manifestCopy.tcpPorts = { 'env': { title: 34 } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad title

        manifestCopy.tcpPorts = { 'env': { title: 't', description: 'long enough' } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // short title

        manifestCopy.tcpPorts = { 'env': { title: 'title', description: 'description', containerPort: 'invalid' } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad containerPort

        manifestCopy.tcpPorts = { 'env': { title: 'title', description: 'description', containerPort: NaN } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad containerPort

        manifestCopy.tcpPorts = { 'env': { title: 'title', description: 'description', containerPort: -2 } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad containerPort

        manifestCopy.tcpPorts = { 'env': { title: 'title', description: 'description', defaultValue: 'invalid' } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad defaultValue

        manifestCopy.tcpPorts = { 'env': { title: 'title', description: 'description', defaultValue: NaN } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad defaultValue

        manifestCopy.tcpPorts = { 'env': { title: 'title', description: 'description', defaultValue: -2 } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad defaultValue
    });

    it('fails for bad udpPorts', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.udpPorts = 45;
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);

        manifestCopy.udpPorts = { 'env$': { } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad env

        manifestCopy.udpPorts = { 'env': { } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // missing description

        manifestCopy.udpPorts = { 'env': { description: 34 } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad description

        manifestCopy.udpPorts = { 'env': { title: 34 } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad title

        manifestCopy.udpPorts = { 'env': { title: 't', description: 'long enough' } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // short title

        manifestCopy.udpPorts = { 'env': { title: 'title', description: 'description', containerPort: 'invalid' } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad containerPort

        manifestCopy.udpPorts = { 'env': { title: 'title', description: 'description', containerPort: NaN } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad containerPort

        manifestCopy.udpPorts = { 'env': { title: 'title', description: 'description', containerPort: -2 } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad containerPort

        manifestCopy.udpPorts = { 'env': { title: 'title', description: 'description', defaultValue: 'invalid' } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad defaultValue

        manifestCopy.udpPorts = { 'env': { title: 'title', description: 'description', defaultValue: NaN } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad defaultValue

        manifestCopy.udpPorts = { 'env': { title: 'title', description: 'description', defaultValue: -2 } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // bad defaultValue
    });

    it('fails for duplicate key in udp/tcpPorts', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.udpPorts = { 'UGOOD_PORT': { title: 't1', description: 'description'}, 'DUP_PORT': { title: 't2', description: 'duplicate' } };
        manifestCopy.tcpPorts = { 'DUP_PORT': { title: 't2', description: 'duplicate' }, 'TGOOD_PORT': { title: 't1', description: 'description'} };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error); // duplicate DUP_PORT
    });

    it('fails for bad addon', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.addons = { 'andy': { } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('fails for bad tags', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.tags = [ 234 ];
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);

        const manifestCopy2 = _.extend({ }, manifest);
        manifestCopy2.tags = [ 'again', 'and', 'again' ]; // duplicate
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy2)).error).to.be.an(Error);
    });

    it('fails for bad scheduler addon', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.addons = {
            scheduler: {
                'nice': 'try'
            }
        };

        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);

        manifestCopy.addons = {
            scheduler: {
                'task1': { schedule: '1 2 3 4 m', command: 'blah' }
            }
        };

        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('succeeds for good scheduler pattern', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.addons = {
            'scheduler': {
                'task1': { schedule: '* * * * *', command: 'blah' }
            }
        };

        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });

    it('succeeds for good scheduler named pattern', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.addons = {
            'scheduler': {
                'task1': { schedule: '@service', command: 'blah' }
            }
        };

        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });

    it('succeeds for good capabilities', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.capabilities = [ ];
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);

        const manifestCopy2 = _.extend({ }, manifest);
        manifestCopy2.capabilities = [ 'net_admin', 'mlock', 'ping' ];
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy2)).manifest).to.eql(manifestCopy2);
    });

    it('succeeds for good tcpPorts', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.tcpPorts = { 'env': { title: '12345', description: '12345', containerPort: 546 } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);

        const manifestCopy2 = _.extend({ }, manifest);
        manifestCopy2.tcpPorts = { 'env': { title: '12345', description: '12345', containerPort: 65535 } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy2)).manifest).to.eql(manifestCopy2);
    });

    it('succeeds for good udpPorts', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.udpPorts = { 'env': { title: '12345', description: '12345', containerPort: 546 } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);

        const manifestCopy2 = _.extend({ }, manifest);
        manifestCopy2.udpPorts = { 'env': { title: '12345', description: '12345', containerPort: 65535 } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy2)).manifest).to.eql(manifestCopy2);
    });

    it('succeeds for minimal valid manifest', function () {
        expect(manifestFormat.parseString(JSON.stringify(manifest)).manifest).to.eql(manifest);
    });

    it('succeeds for maximal valid manifest', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.minBoxVersion = '0.0.1';
        manifestCopy.maxBoxVersion = '1.0.0';
        manifestCopy.targetBoxVersion = '1.0.0';
        manifestCopy.addons = {
            'mysql': { },
            'postgresql': { },
            'mongodb': { cacheSize: 34 }
        };
        manifestCopy.changelog = 'Change upload size limit';
        manifestCopy.tags = [ 'wiki', 'collaboration' ];
        manifestCopy.capabilities = [ 'net_admin' ];

        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });

    it('fails for bad memoryLimit', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.memoryLimit = '1p';
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('succeeds for int memoryLimit', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.memoryLimit = 123456;
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });

    it('succeeds for string memoryLimit', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.memoryLimit = '10MB';
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });


    it('fails for non uri type mediaLinks', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.mediaLinks = [ 'http://foobar.de', 'http://baz', 'foobar' ];
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('fails for duplicate mediaLinks', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.mediaLinks = [ 'http://foobar.de', 'http://baz.com', 'http://foobar.de' ];
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('succeeds with mediaLinks', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.mediaLinks = [ 'http://foobar.de', 'http://baz.com' ];
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });

    it('succeeds with upstreamVersion', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.upstreamVersion = '1.2';
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });

    it('succeeds with postInstallMessage', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.postInstallMessage = 'Can you see this? Thanks for installing';
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });

    it('succeeds with optionalSso', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.optionalSso = true;
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });

    it('fails with invalid documentationUrl', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.documentationUrl = 'not-valid';
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('succeeds with forumUrl', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.forumUrl = 'https://cloudron.io/documentation/apps/wordpress';
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });

    it('fails with invalid forumUrl', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.forumUrl = 'not-valid';
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('succeeds with documentationUrl', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.documentationUrl = 'https://cloudron.io/documentation/apps/wordpress';
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });

    it('succeeds for proxyAuth', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.addons = { proxyAuth: {} };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });

    // deprecated
    it('fails for bad multiDomain', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.multiDomain = { };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    // deprecated
    it('succeeds for multiDomain', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.multiDomain = true;
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });

    it('fails for bad aliasableDomain', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.aliasableDomain = { };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('succeeds for aliasableDomain', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.aliasableDomain = true;
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });

    it('fails with no httpPort or httpPorts', function () {
        const manifestCopy = _.extend({ }, manifest);
        delete manifestCopy.httpPort;
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('succeed with httpPorts', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.httpPorts = { 'UI_DOMAIN': { title: 'title', description: 'description', containerPort: 9000, defaultValue: 'ui.domain', aliasableDomain: true } };
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });

    it('succeed with logPaths', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.logPaths = [ 'path1', 'path2' ];
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });

    it('fails with bad runtimeDirs', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.runtimeDirs = [ '/whatever' ];
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).error).to.be.an(Error);
    });

    it('succeed with runtimeDirs', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.runtimeDirs = [ '/app/code', '/app/code/bar', '/home/cloudron/.npm' ];
        expect(manifestFormat.parseString(JSON.stringify(manifestCopy)).manifest).to.eql(manifestCopy);
    });
});

describe('isId', function () {
    it('return false for invalid ids', function () {
        expect(manifestFormat.isId('combarzoo')).to.be(false);
        expect(manifestFormat.isId('com.bar-zoo')).to.be(false);
        expect(manifestFormat.isId('c.bar.zoo')).to.be(false);
    });

    it('succeeds for valid ids', function () {
        expect(manifestFormat.isId('com.bar.zoo')).to.be(true);
        expect(manifestFormat.isId('com.bar_zoo')).to.be(true);
        expect(manifestFormat.isId('com.barzoo')).to.be(true);
        expect(manifestFormat.isId('co.bar.zoo')).to.be(true);
    });
});

describe('checkAppstoreRequirements', function () {
    const manifest = {
        id: 'io.cloudron.test',
        title: 'Bar App',
        description: 'Long long ago, there was foo',
        tagline: 'Not your usual Foo App',
        author: 'Herbert Burgermeister',
        manifestVersion: 1,
        version: '0.1.2',
        dockerImage: 'cloudron/io.cloudron.test:0.1.2',
        httpPort: 23,
        healthCheckPath: '/',
        website: 'https://example.com',
        contactEmail: 'support@example.com',
        tags: [ 'something' ],
        mediaLinks: [ 'https://foo.ong' ],
        icon: 'file://icon.png',
        changelog: 'file://changelog.md',
        customAuth: false,
        postInstallMessage: 'Nice!',
    };

    it('fails without tags', function () {
        const manifestCopy = _.extend({ }, manifest);
        delete manifestCopy.tags;
        expect(manifestFormat.checkAppstoreRequirements(manifestCopy)).to.be.an(Error);
    });

    it('fails without mediaLinks', function () {
        const manifestCopy = _.extend({ }, manifest);
        delete manifestCopy.mediaLinks;
        expect(manifestFormat.checkAppstoreRequirements(manifestCopy)).to.be.an(Error);
    });

    it('fails without icon', function () {
        const manifestCopy = _.extend({ }, manifest);
        delete manifestCopy.icon;
        expect(manifestFormat.checkAppstoreRequirements(manifestCopy)).to.be.an(Error);
    });

    it('fails without changelog', function () {
        const manifestCopy = _.extend({ }, manifest);
        delete manifestCopy.changelog;
        expect(manifestFormat.checkAppstoreRequirements(manifestCopy)).to.be.an(Error);
    });

    it('fails with prerelease version', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.version = '1.2.3-pre';
        expect(manifestFormat.checkAppstoreRequirements(manifestCopy)).to.be.an(Error);
    });

    it('fails with build version', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.version = '1.2.3+build';
        expect(manifestFormat.checkAppstoreRequirements(manifestCopy)).to.be.an(Error);
    });

    it('succeeds with numeric prerelease version', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.dockerImage = 'cloudron/io.cloudron.test:1.2.3-1-timestamp';
        manifestCopy.version = '1.2.3-1';
        expect(manifestFormat.checkAppstoreRequirements(manifestCopy)).to.be(null);
    });

    it('requires 7.1 atleast for upstreamVersion', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.upstreamVersion = '3.5';
        expect(manifestFormat.checkAppstoreRequirements(manifestCopy)).to.be.an(Error);
        manifestCopy.minBoxVersion = '7.1.0';
        expect(manifestFormat.checkAppstoreRequirements(manifestCopy)).to.be(null);
    });

    it('requires 7.1 atleast for logPaths', function () {
        const manifestCopy = _.extend({ }, manifest);
        manifestCopy.logPaths = [ 'somefile' ];
        expect(manifestFormat.checkAppstoreRequirements(manifestCopy)).to.be.an(Error);
        manifestCopy.minBoxVersion = '7.1.0';
        expect(manifestFormat.checkAppstoreRequirements(manifestCopy)).to.be(null);
    });

    it('succeeds with all requirements', function () {
        const manifestCopy = _.extend({ }, manifest);
        expect(manifestFormat.checkAppstoreRequirements(manifestCopy)).to.be(null);
    });
});

